from bitmerchant.wallet import (
    Wallet as HDWallet
)
from bitcoin import (
    bip32_serialize,
    encode_pubkey as encode_public_key
)
from binascii import hexlify, unhexlify
from .utils import extract_bin_chain_path
from .configs import EXTENDED_PUBLIC_KEY_VERSION_BYTES as version_bytes


class PublicKeychain():
    def __init__(self, public_keychain):
        if isinstance(public_keychain, HDWallet):
            self.hdkeychain = public_keychain
        elif isinstance(public_keychain, (str, unicode)):
            self.hdkeychain = HDWallet.deserialize(public_keychain)
        else:
            raise ValueError('public keychain must be a string')

    def __str__(self):
        return self.hdkeychain.serialize_b58(private=False)

    def child(self, index):
        child_keychain = self.hdkeychain.get_child(
            index, is_prime=False, as_private=False)
        return PublicKeychain(child_keychain)

    def descendant(self, chain_path):
        """ A descendant is a child many steps down.
        """
        public_child = self.hdkeychain
        chain_step_bytes = 4
        max_bits_per_step = 2**31
        chain_steps = [
            int(chain_path[i:i+chain_step_bytes*2], 16) % max_bits_per_step
            for i in range(0, len(chain_path), chain_step_bytes*2)
        ]
        for step in chain_steps:
            public_child = public_child.get_child(step)

        return PublicKeychain(public_child)

    def public_key(self, compressed=True):
        return self.hdkeychain.get_public_key_hex(compressed=compressed)

    def address(self):
        return str(self.hdkeychain.to_address())

    @classmethod
    def from_public_key(cls, public_key, chain_path='\x00'*32, depth=0,
                        fingerprint='\x00'*4, child_index=0):
        public_key_bytes = encode_public_key(public_key, 'bin_compressed')
        chain_path = extract_bin_chain_path(chain_path)
        keychain_parts = (version_bytes, depth, fingerprint,
                          child_index, chain_path, public_key_bytes)
        public_keychain_string = bip32_serialize(keychain_parts)
        return PublicKeychain(public_keychain_string)



from bitmerchant.wallet import Wallet as HDWallet
from .public_keychain import PublicKeychain
from binascii import unhexlify, hexlify
from bitcoin import (
    bip32_serialize,
    encode_privkey as encode_private_key
)
from .utils import extract_bin_chain_path
from .configs import EXTENDED_PRIVATE_KEY_VERSION_BYTES as version_bytes


class PrivateKeychain():
    def __init__(self, private_keychain=None):
        if private_keychain:
            if isinstance(private_keychain, HDWallet):
                self.hdkeychain = private_keychain
            elif isinstance(private_keychain, (str, unicode)):
                self.hdkeychain = HDWallet.deserialize(private_keychain)
            else:
                raise ValueError('private keychain must be a string')
        else:
            self.hdkeychain = HDWallet.new_random_wallet()

    def __str__(self):
        return self.hdkeychain.serialize_b58(private=True)

    def hardened_child(self, index):
        child_keychain = self.hdkeychain.get_child(
            index, is_prime=True, as_private=True)
        return PrivateKeychain(child_keychain)

    def child(self, index):
        child_keychain = self.hdkeychain.get_child(
            index, is_prime=False, as_private=True)
        return PrivateKeychain(child_keychain)

    def public_keychain(self):
        public_keychain = self.hdkeychain.public_copy()
        return PublicKeychain(public_keychain)

    def private_key(self, compressed=True):
        private_key = self.hdkeychain.get_private_key_hex()
        if compressed:
            private_key += '01'
        return private_key

    @classmethod
    def from_private_key(cls, private_key, chain_path='\x00'*32, depth=0,
                         fingerprint='\x00'*4, child_index=0):
        private_key_bytes = encode_private_key(private_key, 'bin_compressed')
        chain_path = extract_bin_chain_path(chain_path)
        keychain_parts = (version_bytes, depth, fingerprint,
                          child_index, chain_path, private_key_bytes)
        public_keychain_string = bip32_serialize(keychain_parts)
        return PrivateKeychain(public_keychain_string)



from binascii import unhexlify


def extract_bin_chain_path(chain_path):
    if len(chain_path) == 64:
        return unhexlify(chain_path)
    elif len(chain_path) == 32:
        return chain_path
    else:
        raise ValueError('Invalid chain path')



"""
keychain
==============

"""

from setuptools import setup, find_packages

setup(
    name='keychain',
    version='0.1.4',
    url='https://github.com/blockstack/keychain-manager-py',
    license='MIT',
    author='Blockstack Developers',
    author_email='hello@onename.com',
    description="""Library for BIP32 hierarchical deterministic keychains / wallets.""",
    keywords='bitcoin blockchain bip32 HD hierarchical deterministic keychain wallet',
    packages=find_packages(),
    zip_safe=False,
    install_requires=[
        'bitmerchant>=0.1.8',
        'bitcoin>=1.1.39'
    ],
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet',
        'Topic :: Security :: Cryptography',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)
